@@ -1083,18 +1083,19 @@ public:
                 return 0; // it means nothing to do for the caller
             }
 
+            reserve_top_blocks(i + 1);
             if (initial_block_type == 0) // bitset requested
             {
                 block = alloc_.alloc_bit_block();
                 // initialize block depending on its previous status
                 bm::bit_block_set(block, block_flag ? ~0u : 0u);
-                set_block(nb, block);
+                set_block(i, j, block, false/*bit*/);
             }
             else // gap block requested
             {
                 bm::gap_word_t* gap_block = allocate_gap_block(0);
-                gap_set_all(gap_block, bm::gap_max_bits, block_flag);
-                set_block(nb, (bm::word_t*)gap_block, true/*gap*/);
+                bm::gap_set_all(gap_block, bm::gap_max_bits, block_flag);
+                set_block(i, j, (bm::word_t*)gap_block, true/*gap*/);
                 return (bm::word_t*)gap_block;
             }
         }
@@ -2175,6 +2176,120 @@ public:
         return s_size;
     }
 
+    // ----------------------------------------------------------------
+
+    void optimize_block(unsigned i, unsigned j,
+                        bm::word_t*    block,
+                        bm::word_t*    temp_block,
+                        int            opt_mode,
+                        bv_statistics* bv_stat)
+    {
+        if (!IS_VALID_ADDR(block))
+            return;
+
+        if (BM_IS_GAP(block)) // gap block
+        {
+            gap_word_t* gap_blk = BMGAP_PTR(block);
+            if (bm::gap_is_all_zero(gap_blk))
+            {
+                set_block_ptr(i, j, 0);
+                alloc_.free_gap_block(gap_blk, glen());
+            }
+            else
+            if (bm::gap_is_all_one(gap_blk))
+            {
+                set_block_ptr(i, j, FULL_BLOCK_FAKE_ADDR);
+                alloc_.free_gap_block(gap_blk, glen());
+            }
+            else
+            {
+                // try to shrink the GAP block
+                //
+                unsigned len = bm::gap_length(gap_blk);
+                int old_level = bm::gap_level(gap_blk);
+                int new_level = bm::gap_calc_level(len, glen());
+                if (new_level >= 0 && new_level < old_level) // can shrink
+                {
+                    gap_word_t* new_gap_blk =
+                        allocate_gap_block(unsigned(new_level), gap_blk);
+                    bm::word_t* blk = (bm::word_t*)BMPTR_SETBIT0(new_gap_blk);
+                    set_block_ptr(i, j, blk);
+                    alloc_.free_gap_block(gap_blk, glen());
+                    gap_blk = new_gap_blk;
+                }
+
+                if (bv_stat)
+                {
+                    bv_stat->add_gap_block(
+                        bm::gap_capacity(gap_blk, glen()), len);
+                }
+            }
+        }
+        else // bit-block
+        {
+            // evaluate bit-block complexity
+            unsigned gap_count = bm::bit_block_calc_change(block);
+            if (gap_count == 1) // all-zero OR all-one block
+            {
+                if ((block[0] & 1u)) // all-ONE
+                {
+                    BM_ASSERT(bm::is_bits_one((wordop_t*)block));
+                    get_allocator().free_bit_block(block);
+                    block = FULL_BLOCK_FAKE_ADDR;
+                }
+                else // empty block
+                {
+                    BM_ASSERT(bm::bit_is_all_zero(block));
+                    get_allocator().free_bit_block(block);
+                    block = 0;
+                }
+                set_block_ptr(i, j, block);
+                return;
+            }
+            if (opt_mode < 3) // free_01 optimization
+            {
+                if (bv_stat)
+                    bv_stat->add_bit_block();
+                return;
+            }
+            // try to compress
+            //
+            gap_word_t* tmp_gap_blk = (bm::gap_word_t*)temp_block;
+            *tmp_gap_blk = bm::gap_max_level << 1;
+            unsigned threashold = glen(bm::gap_max_level)-4;
+            unsigned len = 0;
+            if (gap_count < threashold)
+            {
+                len = bm::bit_to_gap(tmp_gap_blk, block, threashold);
+                BM_ASSERT(len);
+            }
+            if (len)  // GAP compression successful
+            {
+                get_allocator().free_bit_block(block);
+                // check if new gap block can be eliminated
+                BM_ASSERT(!(bm::gap_is_all_zero(tmp_gap_blk)
+                           || bm::gap_is_all_one(tmp_gap_blk)));
+
+                int level = bm::gap_calc_level(len, glen());
+                BM_ASSERT(level >= 0);
+                gap_word_t* gap_blk =
+                    allocate_gap_block(unsigned(level), tmp_gap_blk);
+                bm::word_t* blk = (bm::word_t*)BMPTR_SETBIT0(gap_blk);
+                set_block_ptr(i, j, blk);
+                if (bv_stat)
+                {
+                    bv_stat->add_gap_block(
+                            bm::gap_capacity(gap_blk, glen()),
+                            bm::gap_length(gap_blk));
+                }
+            }
+            else  // non-compressable bit-block
+            {
+                if (bv_stat)
+                    bv_stat->add_bit_block();
+            }
+        } // bit-block
+    }
 
     // ----------------------------------------------------------------
     
@@ -2201,114 +2316,8 @@ public:
             
             for (unsigned j = 0; j < bm::set_sub_array_size; ++j)
             {
-                bm::word_t* block = blk_blk[j];
-                if (IS_VALID_ADDR(block))
-                {
-                    if (BM_IS_GAP(block)) // gap block
-                    {
-                        gap_word_t* gap_blk = BMGAP_PTR(block);
-                        if (bm::gap_is_all_zero(gap_blk))
-                        {
-                            set_block_ptr(i, j, 0);
-                            alloc_.free_gap_block(gap_blk, glen());
-                        }
-                        else
-                        if (bm::gap_is_all_one(gap_blk))
-                        {
-                            set_block_ptr(i, j, FULL_BLOCK_FAKE_ADDR);
-                            alloc_.free_gap_block(gap_blk, glen());
-                        }
-                        else
-                        {
-                            // try to shrink the GAP block
-                            //
-                            unsigned len = bm::gap_length(gap_blk);
-                            int old_level = bm::gap_level(gap_blk);
-                            int new_level = bm::gap_calc_level(len, glen());
-                            if (new_level >= 0 && new_level < old_level) // can shrink
-                            {
-                                gap_word_t* new_gap_blk =
-                                    allocate_gap_block(unsigned(new_level), gap_blk);
-                                bm::word_t* blk = (bm::word_t*)BMPTR_SETBIT0(new_gap_blk);
-                                set_block_ptr(i, j, blk);
-                                alloc_.free_gap_block(gap_blk, glen());
-                                gap_blk = new_gap_blk;
-                            }
-
-                            if (bv_stat)
-                            {
-                                bv_stat->add_gap_block(
-                                        bm::gap_capacity(gap_blk, glen()),
-                                        len);
-                            }
-                        }
-                    }
-                    else // bit-block
-                    {
-                        // evaluate bit-block complexity
-                        unsigned gap_count = bm::bit_block_calc_change(block);
-                        if (gap_count == 1) // all-zero OR all-one block
-                        {
-                            if ((block[0] & 1u)) // all-ONE
-                            {
-                                BM_ASSERT(bm::is_bits_one((wordop_t*)block));
-                                get_allocator().free_bit_block(block);
-                                block = FULL_BLOCK_FAKE_ADDR;
-                            }
-                            else // empty block
-                            {
-                                BM_ASSERT(bm::bit_is_all_zero(block));
-                                get_allocator().free_bit_block(block);
-                                block = 0;
-                            }
-                            set_block_ptr(i, j, block);
-                            continue;
-                        }
-                        if (opt_mode < 3) // free_01 optimization
-                        {
-                            if (bv_stat)
-                                bv_stat->add_bit_block();
-                            continue;
-                        }
-                        // try to compress
-                        //
-                        gap_word_t* tmp_gap_blk = (bm::gap_word_t*)temp_block;
-                        *tmp_gap_blk = bm::gap_max_level << 1;
-                        unsigned threashold = glen(bm::gap_max_level)-4;
-                        unsigned len = 0;
-                        if (gap_count < threashold)
-                        {
-                            len = bm::bit_to_gap(tmp_gap_blk, block, threashold);
-                            BM_ASSERT(len);
-                        }
-                        if (len)  // GAP compression successful
-                        {
-                            get_allocator().free_bit_block(block);
-                            // check if new gap block can be eliminated
-                            BM_ASSERT(!(bm::gap_is_all_zero(tmp_gap_blk)
-                                       || bm::gap_is_all_one(tmp_gap_blk)));
-                            
-                            int level = bm::gap_calc_level(len, glen());
-                            BM_ASSERT(level >= 0);
-                            gap_word_t* gap_blk =
-                                allocate_gap_block(unsigned(level), tmp_gap_blk);
-                            bm::word_t* blk = (bm::word_t*)BMPTR_SETBIT0(gap_blk);
-                            set_block_ptr(i, j, blk);
-                            if (bv_stat)
-                            {
-                                bv_stat->add_gap_block(
-                                        bm::gap_capacity(gap_blk, glen()),
-                                        bm::gap_length(gap_blk));
-                            }
-                        }
-                        else  // non-compressable bit-block
-                        {
-                            if (bv_stat)
-                                bv_stat->add_bit_block();
-                        }
-                    } // bit-block
-                }
-            } // for j
+                optimize_block(i, j, blk_blk[j], temp_block, opt_mode, bv_stat);
+            }
             
             // optimize the top level
             //
@@ -2326,7 +2335,6 @@ public:
                 if (bv_stat)
                     bv_stat->ptr_sub_blocks++;
             }
-
         } // for i
         
         if (bv_stat)
@@ -2336,6 +2344,26 @@ public:
         }
     }
 
+    // ----------------------------------------------------------------
+    
+    void stat_correction(bv_statistics* stat) noexcept
+    {
+        BM_ASSERT(stat);
+        
+        size_t safe_inc = stat->max_serialize_mem / 10; // 10% increment
+        if (!safe_inc) safe_inc = 256;
+        stat->max_serialize_mem += safe_inc;
+
+        unsigned top_size = top_block_size();
+        size_t blocks_mem = sizeof(*this);
+        blocks_mem += sizeof(bm::word_t**) * top_size;
+        blocks_mem += stat->ptr_sub_blocks * (sizeof(void*) * bm::set_sub_array_size);
+        stat->memory_used += blocks_mem;
+        stat->bv_count = 1;
+    }
+
+
+
     // ----------------------------------------------------------------
 
     /// Allocate arena (content memory) based on arena statistics
