@@ -211,6 +211,7 @@ public:
         
         this->size_ = new_size;
     }
+
     /// adjust current size (no need to reallocate)
     void resize_no_check(size_t new_size) BMNOEXCEPT
     {
@@ -371,7 +372,7 @@ public:
         }
     }
     
-    value_type* data() BMNOEXCEPT { return (value_type*) buffer_.data(); }
+    value_type* data() const BMNOEXCEPT { return (value_type*) buffer_.data(); }
 
     void swap(heap_vector<Val, BVAlloc, trivial_type>& other) BMNOEXCEPT
     {
@@ -449,7 +450,7 @@ public:
             return;
         if (new_size < sz) // shrink
         {
-            if (!trivial_type)
+            if constexpr (!trivial_type)
             {
                 unsigned char* this_data = buffer_.data();
                 for (size_type i = new_size; i < sz; ++i)
@@ -463,7 +464,7 @@ public:
         else
         {
             buffer_.resize(new_size * v_size);
-            if (!trivial_type)
+            if constexpr (!trivial_type)
             {
                 unsigned char* this_data = buffer_.data();
                 for (size_type i = sz; i < new_size; ++i)
