@@ -76,7 +76,8 @@ public:
     typedef typename allocator_type::allocator_pool_type allocator_pool_type;
     typedef bm::basic_bmatrix<BV>                    bmatrix_type;
     typedef base_sparse_vector<CharType, BV, STR_SIZE> parent_type;
-    
+    typedef typename parent_type::unsigned_value_type unsigned_value_type;
+
     /*! Statistical information about  memory allocation details. */
     struct statistics : public bv_statistics
     {};
@@ -91,7 +92,7 @@ public:
      */
     typedef bm::dynamic_heap_matrix<unsigned char, allocator_type>
                                             slice_octet_matrix_type;
-typedef slice_octet_matrix_type remap_matrix_type;
+    typedef slice_octet_matrix_type remap_matrix_type;
 
     /** Matrix of character frequencies (for optimal code remap)
         @internal
@@ -274,7 +275,7 @@ typedef slice_octet_matrix_type remap_matrix_type;
 
 
         /// \brief Get current position (value)
-        const value_type* value() const BMNOEXCEPT;
+        const value_type* value() const;
 
         /// \brief Get NULL status
         bool is_null() const BMNOEXCEPT { return sv_->is_null(this->pos_); }
@@ -797,11 +798,14 @@ public:
     /*! @name Various traits                                     */
     ///@{
     
-    /** \brief trait if sparse vector is "compressed" (false)
+    /** \brief various type traits
     */
-    static
+    static constexpr
     bool is_compressed() BMNOEXCEPT { return false; }
 
+    static constexpr
+    bool is_str() BMNOEXCEPT { return true; }
+
     ///@}
 
     // ------------------------------------------------------------
@@ -956,7 +960,7 @@ public:
         {
             // can be negative (-1) when bv base offset = 0 and sv = 1,2..
             size_type base = bv_offset - sv_off_;
-            value_type m = mask_;
+            unsigned_value_type m = mask_;
             const unsigned i = substr_i_;
             for (unsigned j = 0; j < bits_size; ++j)
             {
@@ -968,7 +972,7 @@ public:
         void add_range(size_type bv_offset, size_type sz) BMNOEXCEPT
         {
             auto base = bv_offset - sv_off_;
-            value_type m = mask_;
+            unsigned_value_type m = mask_;
             const unsigned i = substr_i_;
             for (size_type j = 0; j < sz; ++j)
             {
@@ -979,7 +983,7 @@ public:
         }
 
         CharMatrix&            cmatr_;     ///< target array for reverse transpose
-        value_type             mask_;      ///< bit-plane mask
+        unsigned_value_type             mask_;      ///< bit-plane mask
         unsigned               substr_i_;  ///< i
         size_type              sv_off_;    ///< SV read offset
     };
@@ -1016,7 +1020,7 @@ public:
                 if (!bv)
                     continue;
 
-                func.mask_ = value_type(1u << bi);
+                func.mask_ = unsigned_value_type(1u << bi);
                 func.sv_off_ = idx_from;
 
                 size_type end = idx_from + dec_size;
@@ -1163,8 +1167,7 @@ protected:
                          bool set_not_null = true)
     {
         BM_ASSERT (cmatr.is_init());
-        //BM_ASSERT (cmatr.rows() >= BufSize);
-        
+
         unsigned max_str_size = 0;
         {
             for (unsigned j = 0; j < imp_size; ++j)
@@ -1185,9 +1188,10 @@ protected:
                     {
                         unsigned char remap_value =
                             remap_matrix2_.get(i, (unsigned char)(ch));
-                        BM_ASSERT(remap_value);
+                        BM_ASSERT(remap_value); // unknown ?!
+                        /*
                         if (!remap_value) // unknown dictionary element
-                            throw_bad_value(0);
+                            throw_bad_value(0); */
                         str[i] = CharType(remap_value);
                     }
                 } // for i
@@ -1196,42 +1200,38 @@ protected:
 
         this->bmatr_.allocate_rows((1+max_str_size) * 8 + this->is_nullable());
 
-        size_type bit_list[BufSize];
+        unsigned_value_type ch_slice[BufSize];
         for (unsigned i = 0; i < max_str_size; ++i)
         {
-            for (unsigned bi = 0; bi < 8; ++bi)
+            unsigned ch_acc = 0;
+#if defined(BMVECTOPT) || defined(BM_USE_GCC_BUILD)
+            if (imp_size == ins_buf_size) /// full buffer import can use loop unrolling
+            {
+                for (size_type j = 0; j < imp_size; j+=4)
+                {
+                    unsigned_value_type ch0 = (unsigned_value_type)cmatr.row(j)[i];
+                    unsigned_value_type ch1 = (unsigned_value_type)cmatr.row(j+1)[i];
+                    unsigned_value_type ch2 = (unsigned_value_type)cmatr.row(j+2)[i];
+                    unsigned_value_type ch3 = (unsigned_value_type)cmatr.row(j+3)[i];
+
+                    ch_acc |= ch0 | ch1 | ch2 | ch3;
+                    ch_slice[j] = ch0; ch_slice[j+1] = ch1;
+                    ch_slice[j+2] = ch2; ch_slice[j+3] = ch3;
+                }
+            }
+            else
+#endif
             {
-                unsigned n_bits = 0;
-                value_type mask = value_type(1u << bi);
                 for (size_type j = 0; j < imp_size; ++j)
                 {
-                    typename CharMatrix::value_type* str = cmatr.row(j);
-                    value_type ch = str[i];
-                    if (!ch)
-                        continue;
-                    if (ch & mask)
-                    {
-                        bit_list[n_bits++] = idx_from + j;
-                        str[i] ^= mask;
-                    }
-                } // for j
-                if (n_bits) // set transposed bits to the target plane
-                {
-                    unsigned plane = i*8 + bi;
-                    /*
-                    bvector_type* bv = this->bmatr_.get_row(plane);
-                    if (!bv)
-                    {
-                        bv = this->bmatr_.construct_row(plane);
-                        bv->init();
-                    }
-                    */
-                    bvector_type* bv = this->get_create_slice(plane);
-                    bv->set(&bit_list[0], n_bits, BM_SORTED);
+                    unsigned_value_type ch = (unsigned_value_type)cmatr.row(j)[i];
+                    ch_acc |= ch;
+                    ch_slice[j] = ch;
                 }
-            } // for k
-        } // for i
-        
+            }
+            import_char_slice(ch_slice, ch_acc, i, idx_from, imp_size);
+        }
+
         size_type idx_to = idx_from + imp_size - 1;
         if (set_not_null)
         {
@@ -1243,6 +1243,62 @@ protected:
 
     }
 
+    /// @internal
+    template<size_t BufSize = ins_buf_size>
+    void import_char_slice(const unsigned_value_type* ch_slice,
+                           unsigned ch_acc,
+                           size_type char_slice_idx,
+                           size_type idx_from, size_type imp_size)
+    {
+        size_type bit_list[BufSize];
+        for ( ;ch_acc; ch_acc &= ch_acc - 1) // bit-scan
+        {
+            unsigned n_bits = 0;
+            const unsigned bi = (bm::word_bitcount((ch_acc & -ch_acc) - 1));
+            unsigned mask = 1u << bi;
+#if defined(BMVECTOPT) || defined(BM_USE_GCC_BUILD)
+            if (imp_size == ins_buf_size) /// full buffer import can use loop unrolling
+            {
+                mask |= (mask << 8) | (mask << 16) | (mask << 24);
+                for (size_type j = 0; j < imp_size; j+=4)
+                {
+                    unsigned ch0 = ((unsigned)ch_slice[j+0]) |
+                                   ((unsigned)ch_slice[j+1] << 8)  |
+                                   ((unsigned)ch_slice[j+2] << 16) |
+                                   ((unsigned)ch_slice[j+3] << 24);
+                    ch0 &= mask;
+                    ch0 = (ch0 >> bi) | (ch0 >> (bi+7)) |
+                          (ch0 >> (bi+14)) | (ch0 >> (bi+21));
+                    ch0 &= 15u;
+                    BM_ASSERT(bm::word_bitcount(ch0) <= 4);
+                    for (size_type base_idx = idx_from + j ;ch0; ch0 &= ch0 - 1) // bit-scan
+                    {
+                        const unsigned bit_idx =
+                                (bm::word_bitcount((ch0 & -ch0) - 1));
+                        bit_list[n_bits++] = base_idx + bit_idx;
+                    } // for ch0
+                } // for j
+            }
+            else
+#endif
+            {
+                for (size_type j = 0; j < imp_size; ++j)
+                {
+                    unsigned ch = unsigned(ch_slice[j]);
+                    if (ch & mask)
+                        bit_list[n_bits++] = idx_from + j;
+                } // for j
+            }
+
+            if (n_bits) // set transposed bits to the target plane
+            {
+                bvector_type* bv =
+                    this->get_create_slice((unsigned)(char_slice_idx * 8) + bi);
+                bv->import_sorted(&bit_list[0], n_bits);
+            }
+        } // for ch_acc
+    }
+
     // ------------------------------------------------------------
     /*! @name Errors and exceptions                              */
     ///@{
@@ -1774,9 +1830,7 @@ bool str_sparse_vector<CharType, BV, STR_SIZE>::remap_tosv(
         const unsigned char* remap_row = octet_remap_matrix2.row(i);
         unsigned char remap_value = remap_row[unsigned(ch)];
         if (!remap_value) // unknown dictionary element
-        {
             return false;
-        }
         sv_str[i] = CharType(remap_value);
     } // for i
     return true;
@@ -2104,7 +2158,7 @@ void str_sparse_vector<CharType, BV, STR_SIZE>::const_iterator::set_substr(
 
 template<class CharType, class BV, unsigned STR_SIZE>
 const typename str_sparse_vector<CharType, BV, STR_SIZE>::value_type*
-str_sparse_vector<CharType, BV, STR_SIZE>::const_iterator::value() const BMNOEXCEPT
+str_sparse_vector<CharType, BV, STR_SIZE>::const_iterator::value() const
 {
     BM_ASSERT(sv_);
     BM_ASSERT(this->valid());
