@@ -86,6 +86,7 @@ bm::chrono_taker::duration_map_type     timing_map;
 
 
 typedef bm::bvector<> bvect;
+typedef bvect::allocator_type  bv_allocator_type;
 
 typedef std::vector<unsigned> ref_vect_type;
 
@@ -96,9 +97,9 @@ typedef bm::rsc_sparse_vector<unsigned long long, sparse_vector_u64> rsc_sparse_
 
 
 typedef
-bm::thread_pool<bm::task_description*, bm::spin_lock<bm::pad0_struct> > pool_spin_type;
+bm::thread_pool<bm::task_descr*, bm::spin_lock<bm::pad0_struct> > pool_spin_type;
 typedef
-bm::thread_pool<bm::task_description*, std::mutex > pool_mutex_type;
+bm::thread_pool<bm::task_descr*, std::mutex > pool_mutex_type;
 
 
 
@@ -163,6 +164,49 @@ void TestSimpleQueue()
     cout << " ----------------------------- TestSimpleQueue() OK" << endl;
 }
 
+static
+void TestTaskDescr()
+{
+    cout << " ----------------------------- TestTaskDescr()" << endl;
+
+    {
+        task_descr td;
+    }
+    {
+        std::function<int(void*)> f1([](void*)
+        {
+            return 10;
+        });
+    }
+
+    {
+        task_descr td([](void*) { return 2; }, nullptr);
+        int ret = td.func(td.argp);
+        assert(ret == 2);
+    }
+
+    {
+        task_descr td;
+        td.init([](void*) { return 3; }, nullptr);
+        td.run();
+        assert(td.done == 1);
+        assert(td.err_code = 3);
+    }
+
+    {
+        typedef bm::task_batch<bv_allocator_type> task_batch_type;
+        task_batch_type tb;
+        std::function<int(void*)> f1([](void*) { return 110; });
+        tb.add(f1, nullptr);
+        assert(tb.size() == 1);
+        int ret = tb.get_task(0)->run();
+        assert(ret == 110);
+    }
+
+    cout << " ----------------------------- TestTaskDescr() OK" << endl;
+}
+
+
 
 static
 void Check_SimModel(const bm::xor_sim_model<bvect>& sim1,
@@ -245,7 +289,7 @@ void generate_df_svu32_set(sparse_vector_u32& sv1i,
         bi2 = 0;
         bi3 = 1 + 0;
         bi3.add_null();
-        bi4 = rand() % 256000;
+        bi4 = (bvect::size_type)rand() % 256000;
         bi4.add_null();
 
     } // for
@@ -257,6 +301,7 @@ void generate_df_svu32_set(sparse_vector_u32& sv1i,
 
 }
 
+
 template<typename PoolType>
 void TestParallelSV_Serial(const char* test_label,
                            bool        check_sim,
@@ -291,6 +336,9 @@ void TestParallelSV_Serial(const char* test_label,
         pool_type tpool;  // our thread pool here (no threads created yet)
         tpool.start(num_threads); // start the threads
 
+        bm::serializer<bvect>::xor_sim_model_type sim_model_c;
+        bool control_ready(false);
+
         bm::chrono_taker tt1(test_label, 1, &timing_map);
 
         for (unsigned pass = 0; pass < 2; ++pass)
@@ -307,6 +355,13 @@ void TestParallelSV_Serial(const char* test_label,
 
             bm::sparse_vector_serializer<sparse_vector_u32>::xor_sim_model_type sim_model;
 
+            if (check_sim && !control_ready)
+            {
+                sv_serializer.compute_sim_model(sim_model_c, bv_ref, xor_search_params);
+                control_ready = true; // to avoid double re-calc
+            }
+
+
             {
                 bm::compute_sim_matrix_plan_builder<bvect> pbuilder;
                 bm::compute_sim_matrix_plan_builder<bvect>::task_batch tbatch;
@@ -318,12 +373,12 @@ void TestParallelSV_Serial(const char* test_label,
                     bm::thread_pool_executor<pool_type> exec;
                     exec.run(tpool, tbatch, true);
                 }
-                
+
                 if (check_sim)
                 {
-                    bm::serializer<bvect>::xor_sim_model_type sim_model_c;
-                    sv_serializer.compute_sim_model(sim_model_c, bv_ref, xor_search_params);
+                    assert(control_ready);
                     Check_SimModel(sim_model_c, sim_model);
+                    cout << "XOR model check - OK." << endl;
                 }
             }
 
@@ -335,10 +390,12 @@ void TestParallelSV_Serial(const char* test_label,
                 sbuilder.set_xor_ref(&bv_ref);
                 sbuilder.set_sim_model(&sim_model);
 
+                /*
                 {
                     bm::sv_serialization_plan_builder<sparse_vector_u32>::task_batch tbatch;
                     sbuilder.build_plan(tbatch, sv_lay1p, sv1i);
                 }
+                */
 
                 sv_serializer.serialize(sv1i, sv_lay1);
                 {
@@ -364,6 +421,8 @@ void TestParallelSV_Serial(const char* test_label,
                 }
             }
 
+            sim_model.bv_blocks.clear(true);
+
             // ----------
 
             if (deserialize)
@@ -414,10 +473,35 @@ void TestParallelSV_Serial(const char* test_label,
 
             if (pass == 0)
             {
-                sv1i.optimize();
-                sv2i.optimize();
-                sv3i.optimize();
-                sv4i.optimize();
+                cout << " sparse vector optimization... " << flush;
+
+                bm::optimize_plan_builder<sparse_vector_u32, std::mutex> op_builder;
+                bm::optimize_plan_builder<sparse_vector_u32, std::mutex>::task_batch opt_batch;
+
+
+                op_builder.build_plan(opt_batch, sv1i);
+                auto bsize = opt_batch.size();
+                assert(bsize > 0);
+
+                op_builder.build_plan(opt_batch, sv2i);
+                assert(bsize < opt_batch.size());
+                bsize = opt_batch.size();
+
+                op_builder.build_plan(opt_batch, sv3i);
+                assert(bsize < opt_batch.size());
+                bsize = opt_batch.size();
+
+                op_builder.build_plan(opt_batch, sv4i);
+                assert(bsize < opt_batch.size());
+
+                // run the plan
+                {
+                    bm::thread_pool_executor<pool_type> exec;
+                    exec.run(tpool, opt_batch, true);
+                }
+
+                cout << "OK" << endl;
+
             }
 
         } // for pass
@@ -429,6 +513,95 @@ void TestParallelSV_Serial(const char* test_label,
     cout << " ----------------------------- TestParallelSV_Serial() OK" << endl;
 }
 
+template<typename PoolType>
+void TestParallelSV_Optimize(const char* test_label)
+{
+    cout << " ----------------------------- TestParallelSV_Optimize() - "
+         << test_label << endl;
+    {
+        sparse_vector_u32 sv1i, sv2i, sv3i(bm::use_null), sv4i(bm::use_null);
+
+        typedef PoolType pool_type;
+
+        {
+            cout << " Sample set generation..." << flush;
+
+            const bvect::size_type idmax = 100000000;
+            generate_df_svu32_set(sv1i, sv2i, sv3i, sv4i, idmax);
+
+            cout << "Ok" << endl;
+        }
+
+        sparse_vector_u32 sv1c(sv1i), sv2c(sv2i), sv3c(sv3i), sv4c(sv4i); // control copy
+        sparse_vector_u32::statistics st1, st2, st3, st4;
+        sparse_vector_u32::statistics st1c, st2c, st3c, st4c;
+
+        {
+            cout << " compute controls... " << flush;
+
+            BM_DECLARE_TEMP_BLOCK(tb)
+
+            std::string l(" (non-parallel)");
+            bm::chrono_taker tt1(test_label + l, 1, &timing_map);
+
+            sv1c.optimize(tb, sparse_vector_u32::bvector_type::opt_compress, &st1c);
+            sv2c.optimize(tb, sparse_vector_u32::bvector_type::opt_compress, &st2c);
+            sv3c.optimize(tb, sparse_vector_u32::bvector_type::opt_compress, &st3c);
+            sv4c.optimize(tb, sparse_vector_u32::bvector_type::opt_compress, &st4c);
+            cout << "OK" << endl;
+        }
+
+        pool_type tpool;  // our thread pool here (no threads created yet)
+        tpool.start(num_threads); // start the threads
+
+        {
+            bm::chrono_taker tt1(test_label, 1, &timing_map);
+
+            cout << " sparse vector optimization... " << flush;
+
+            bm::optimize_plan_builder<sparse_vector_u32, std::mutex> op_builder;
+            bm::optimize_plan_builder<sparse_vector_u32, std::mutex>::task_batch opt_batch;
+
+
+            op_builder.build_plan(opt_batch, sv1i, sparse_vector_u32::bvector_type::opt_compress, &st1);
+            auto bsize = opt_batch.size();
+            assert(bsize > 0);
+
+            op_builder.build_plan(opt_batch, sv2i, sparse_vector_u32::bvector_type::opt_compress, &st2);
+            assert(bsize < opt_batch.size());
+            bsize = opt_batch.size();
+
+            op_builder.build_plan(opt_batch, sv3i, sparse_vector_u32::bvector_type::opt_compress, &st3);
+            assert(bsize < opt_batch.size());
+            bsize = opt_batch.size();
+
+            op_builder.build_plan(opt_batch, sv4i, sparse_vector_u32::bvector_type::opt_compress, &st4);
+            assert(bsize < opt_batch.size());
+
+            // run the plan
+            {
+                bm::thread_pool_executor<pool_type> exec;
+                exec.run(tpool, opt_batch, true);
+            }
+
+            assert(st1.gap_blocks == st1c.gap_blocks);
+            assert(st1.bit_blocks == st1c.bit_blocks);
+            assert(st2.gap_blocks == st2c.gap_blocks);
+            assert(st2.bit_blocks == st2c.bit_blocks);
+            assert(st3.gap_blocks == st3c.gap_blocks);
+            assert(st3.bit_blocks == st3c.bit_blocks);
+            assert(st4.gap_blocks == st4c.gap_blocks);
+            assert(st4.bit_blocks == st4c.bit_blocks);
+
+            cout << "OK" << endl;
+        }
+
+        tpool.set_stop_mode(pool_type::stop_when_done);
+        tpool.join();
+    }
+    cout << " ----------------------------- TestParallelSV_Optimize() OK" << endl;
+}
+
 static
 void show_help()
 {
@@ -488,17 +661,6 @@ int main(int argc, char *argv[])
             cout << "func-type2 " << a << " " << b << endl; return 0;
         });
     cout << sizeof(f2) << endl;
-/*
-    std::vector<job_func_type> job_vect;
-    job_vect.push_back(f1);
-    job_vect.push_back(f2);
-
-    for (size_t i = 0; i < job_vect.size(); ++i)
-    {
-        int r = job_vect[i]();
-        cout << "returned: " << r << endl;
-    }
-*/
     std::deque<job_func_type> job_queue;
     job_queue.push_back(f1);
     job_queue.push_back(f2);
@@ -511,13 +673,6 @@ int main(int argc, char *argv[])
     } // while
 
 
-//    f1();
-//    f2();
-
-
-
-    return 0;
-
 
     {
         auto ret = parse_args(argc, argv);
@@ -534,6 +689,11 @@ int main(int argc, char *argv[])
 
         TestSimpleQueue();
 
+        TestTaskDescr();
+
+        TestParallelSV_Optimize<pool_spin_type>("005-s. optimize() ");
+        TestParallelSV_Optimize<pool_mutex_type>("005-m. optimize() ");
+
 
         TestParallelSV_Serial<pool_spin_type>(
             "001-s. XOR filter (check)", true, false, false);
@@ -548,7 +708,6 @@ int main(int argc, char *argv[])
             "004-s. XOR filter/serialization/deserialization (no-check)",
             false, true, true);
 
-
         TestParallelSV_Serial<pool_mutex_type>(
             "001-m. XOR filter (check)", true, false, false);
 
